//
//  OCENavLineObject.m
//  InfiniteCanvas
//
//  Created by Markus Mueller <markus@mindnode.com>.
//	http://www.infinitecanvas.com/
//  Copyright (c) 2003-2005 Markus Mueller. All rights reserved.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//  

#import "OCENavTrailObject.h"
#import "OCERectObject.h"
#import "OCENavFrameObject.h"
#import "OCEActiveComponent.h"

@implementation OCENavTrailObject

- (id)init 
{
    [super init];
	[self setColor: [NSColor blueColor]];
	allowFreeEndPoints = YES;
	[self setSnapAble:YES];
	transition = 1; // movement
	transitionSpeed = 300;
	transitionColor = [[NSColor whiteColor] retain];
	easing = true;
	
    return self;
}

- (bool) valid
{
	int count = [path elementCount];
	if (count > 1 /*&& startRect && endRect REMOVED IC1.2*/) {
		if (!((count == 2) && (startRect == endRect))) {
			return YES;
		}
	} 
	return NO;
}

- (bool) validForNavigation {
	// valid only checks if it is a valid trail, but not if the trail can be used for navigation
	int count = [path elementCount];
	if (count > 1 && startRect && endRect) {
		if (!((count == 2) && (startRect == endRect))) {
			return YES;
		}
	} 
	return NO;
}

- (void) adaptPathToRects
// overwrite super because we don't clip
{
	if (startRect)
		[self setStartPoint:[startRect center]];
	
	if (endRect)
		[self setEndPoint:[endRect center]];
}

- (void) setAssociatedStartRect: (OCERectObject *) rectObject
{
	[super setAssociatedStartRect: rectObject];
	[self setColor:[(OCENavFrameObject *)rectObject borderColor]];
	// if we change rect, the path should addaped to the color of the new frame
}

- (int) transition
{
	return transition;
}

- (void) setTransition: (int) flag
{
	if (transition != flag)
	{
		[[undoManager prepareWithInvocationTarget:self] setTransition:transition]; 
		[undoManager setActionName: @"Change Transition Type"];

		transition = flag;
	}
}

- (int) transitionSpeed
{
	return transitionSpeed;
}

- (void) setTransitionSpeed: (int) flag
{
	if (transitionSpeed != flag)
	{
		[[undoManager prepareWithInvocationTarget:self] setTransitionSpeed:transitionSpeed]; 
		[undoManager setActionName: @"Change Transition Speed"];

		transitionSpeed = flag;
	}
}

- (NSColor *) transitionColor;
{
	return transitionColor;
}

- (void) setTransitionColor: (NSColor *) newColor;
{	
	[[undoManager prepareWithInvocationTarget:self] setTransitionColor: transitionColor];
	[undoManager setActionName: @"Change Transition Color"];
    
    id old = transitionColor;
    transitionColor = [newColor retain];
    [old release];
}

- (bool) easing;
{
	return easing;
}

- (void) setEasing: (bool) flag;
{
	[[undoManager prepareWithInvocationTarget:self] setEasing: easing];
	[undoManager setActionName: @"Change Use Easing"];

	easing = flag;
}



// Navigation
- (int) segments
{
	return [path elementCount]-1;
}

- (NSPoint) endPointOfSegment: (int) segment
{
	if ((segment < 0) || (segment >= [path elementCount]))
		return NSZeroPoint;
	
	NSPoint *points = malloc(3*sizeof(NSPoint));
	NSPoint returnPoint;
	NSBezierPathElement element = [path elementAtIndex: segment associatedPoints:points];
	
	if (element == NSCurveToBezierPathElement)
		returnPoint = points[2];
	else
		returnPoint = points[0];
	
	free(points);
	
	return returnPoint;
}

- (int) length
{
	int count = [path elementCount];
	int i = 1;
	int length = 0;
	
	while (i < count)
	{
		length += [self lengthOfSegment:i];
		i++;
	}
	return length;
}

- (int) lengthOfSegment: (int) segment
{
	NSPoint startPoint;
	if ((segment-1 == 0) && ([startRect isActiveComponent]))
		startPoint = [[(OCEActiveComponent *)startRect frame] center];
	else
		startPoint = [self endPointOfSegment: segment-1];
	
	NSPoint endPoint = [self endPointOfSegment: segment];
	float width = endPoint.x - startPoint.x;
	float height = endPoint.y - startPoint.y;

	return abs(sqrt(width*width+height*height)); //pythagoras
}

// ----------------------------------------------------------------------
// NSCopying Protocol
// ----------------------------------------------------------------------
- (id)copyWithZone:(NSZone *)zone
{
    OCENavTrailObject *copy = [super copyWithZone: zone];

    copy->transition = transition;
    copy->transitionSpeed = transitionSpeed;
	copy->startRect = startRect;
	copy->endRect = endRect;
	copy->transitionColor = [transitionColor retain];
	copy->easing = easing;

    return copy;
}

// ----------------------------------------------------------------------
// Save and Load
// ----------------------------------------------------------------------
- (NSMutableArray *)dataRepresentation
{
	NSMutableArray *array = [[super dataRepresentation] retain];
	
	[array addObject:[NSNumber numberWithInt:transition]];
	[array addObject:[NSNumber numberWithInt:transitionSpeed]];
	
	NSColor *tmpColor = transitionColor;
	if (![[tmpColor colorSpaceName] isEqualToString:@"NSCalibratedRGBColorSpace"])
		if ((tmpColor = [tmpColor colorUsingColorSpaceName: @"NSCalibratedRGBColorSpace"]) == nil)
			tmpColor = [[NSColor blackColor] colorUsingColorSpaceName: @"NSCalibratedRGBColorSpace"];
    [array addObject: [NSNumber numberWithFloat:[tmpColor redComponent]]];
    [array addObject: [NSNumber numberWithFloat:[tmpColor greenComponent]]];
    [array addObject: [NSNumber numberWithFloat:[tmpColor blueComponent]]];
	[array addObject: [NSNumber numberWithBool: easing]]; // added in 1.3

	return [array autorelease];
}

- (id)initWithContentsOfArray:(NSArray *) array forDocument: (OCEDocument *) newDocument
{
    [self init];
	transition = [[array objectAtIndex:10] intValue];
	transitionSpeed = [[array objectAtIndex:11] intValue];

	if ([array count] > 12) { // new in 1.3
		NSLog(@"testing: load Transition Color");
		[transitionColor release];
		transitionColor = [[NSColor colorWithCalibratedRed: [[array objectAtIndex: 12] floatValue]
												green: [[array objectAtIndex: 13] floatValue]
												blue: [[array objectAtIndex: 14] floatValue] 
												alpha: 1.0] retain];
		easing = [[array objectAtIndex: 15] boolValue];
	} else {
		// we get here if we are loading an old save file
		transitionSpeed *= 100;
		
		switch (transition) {
			case 2:
				transition = 1;
				break;
			case 4:
			case 5:
			case 8:
				transition = 2;
				[self setTransitionColor: [NSColor whiteColor]];
				break;
			case 6:
				transition = 2;
				[self setTransitionColor: [NSColor blackColor]];
				break;
		}
	}
	
    return [super initWithContentsOfArray: array forDocument: newDocument];
}

- (NSString *) exportFlash;
{
	if ([self validForNavigation]) {
		NSMutableString *data = [NSMutableString stringWithFormat:@"<trail type=\""];
		
		switch (transition) {
			case 0: 
				[data appendString: @"none\" "];
				break;
			case 1:
				[data appendString: @"movement\" "];
				break;
			default:
				[data appendString: @"fade\" "];
				NSColor *tmpColor = transitionColor;
				if (![[tmpColor colorSpaceName] isEqualToString:@"NSCalibratedRGBColorSpace"])
					if ((tmpColor = [tmpColor colorUsingColorSpaceName: @"NSCalibratedRGBColorSpace"]) == nil)
						tmpColor = [[NSColor blackColor] colorUsingColorSpaceName: @"NSCalibratedRGBColorSpace"];
				int bgR = [tmpColor redComponent] * 255;
				int bgG = [tmpColor greenComponent] * 255;
				int bgB = [tmpColor blueComponent] * 255;

				[data appendFormat: @"color=\"0x%02x%02x%02x\" ", bgR, bgG, bgB];
				break;
		}
		
		[data appendFormat: @"target=\"%i\" speed=\"%i\"", [endRect rectID], transitionSpeed];
		
		if (easing) {
			[data appendString: @" easing=\"easeInOutQuad\""];
		} 
		
		[data appendString: @">\n"];

		[super exportFlash: data];
		[data appendString:@"</trail>\n"];
		return data;
	} else {
		return @"";
	}
}

@end
