//
//  OCENavFrameObject.m
//  InfiniteCanvas
//
//  Created by Markus Mueller <markus@mindnode.com>.
//	http://www.infinitecanvas.com/
//  Copyright (c) 2003-2005 Markus Mueller. All rights reserved.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//  

#import "OCERectObject.h"
#import "OCEPathObject.h"
#import "OCECanvas.h"
#import "NSBezierPathHitTest.h"
#import "OCEMath.h"
#import "OCEDocument.h"

@implementation OCERectObject

// ----------------------------------------------------------------------
// init Methods
// ----------------------------------------------------------------------
- (id)init
{
    [super init];

    xPos = 0;
    yPos = 0;
    width = 0;
    height = 0;
    rotation = 0;
	proportion = 1.0; // = 1:1
    offsetRect = NSZeroRect;
    isSquare = false;
	rectID = 0;
    
    transform = [[NSAffineTransform transform] retain];
    associatedPaths = [[NSMutableArray arrayWithCapacity:0] retain];
    
    int i;
    for (i=0;i<5;i++)
		saveUndoValuesArray[i] = 0;
    
    return self;
}

- (id) initWithRect: (NSRect) aRect
{
    [self init];
    xPos = aRect.origin.x+aRect.size.width/2.0;
    yPos = aRect.origin.y+aRect.size.height/2.0;
    width = aRect.size.width;
    height = aRect.size.height;
    [self reset];
    
    return self;
}

- (void)dealloc 
{
    [transform release];
    [associatedPaths release];
    [super dealloc];
}

// ----------------------------------------------------------------------
// Draw
// ----------------------------------------------------------------------
- (void) drawObjectWithFraction: (float) fraction isSelected: (bool) selected
{
	[transform concat];
	float drawWidth = ((width == 0) ? 1 : width);
	float drawHeight = ((height == 0) ? 1 : height);
	
	
    NSBezierPath *drawPath = [NSBezierPath bezierPathWithRect: NSMakeRect(-drawWidth/2.0,-drawHeight/2.0,drawWidth,drawHeight)];
    if (fraction == 1)
		[[NSColor blackColor] set];
	else
		[[[NSColor blackColor] colorWithAlphaComponent:fraction] set];
    [drawPath fill];
    
    if (selected)
        [self drawKnobs];
}

- (void) drawKnobs
{
    float length = 6;
    if ((abs(width) < 16) || (abs(height) < 16))
    {
        length = 4;   
    }
    
    // 1. shadow
    // lower left corner
    NSRect rect = NSMakeRect((-width/2.0-length/2.0)+1, (-height/2.0-length/2.0)-1,length,length);
    NSBezierPath *selectPath = [NSBezierPath bezierPathWithRect:rect];
    [selectPath setLineWidth:1.0];
    rect.origin.y += height/2.0; // middle left knob
    if (!isSquare)
        [selectPath appendBezierPathWithRect:rect];
    rect.origin.y += height/2.0; // upper left knob
    [selectPath appendBezierPathWithRect:rect];
    rect.origin.x += width/2.0; // upper middle knob
    if (!isSquare)
        [selectPath appendBezierPathWithRect:rect];
    rect.origin.x += width/2.0; // upper right knob
    [selectPath appendBezierPathWithRect:rect];
    rect.origin.y -= height/2.0; // middle right knob
    if (!isSquare)
        [selectPath appendBezierPathWithRect:rect];
    rect.origin.y -= height/2.0; //lower right knob
    [selectPath appendBezierPathWithRect:rect];
    rect.origin.x -= width/2.0; // lower middle knob
    if (!isSquare)
        [selectPath appendBezierPathWithRect:rect];
    
    [[NSColor blackColor] set];
    [selectPath fill];
    
    // 2. knob
    NSAffineTransform *knobTransform = [NSAffineTransform transform];
    [knobTransform translateXBy: -1 yBy: 1];
    [knobTransform concat];
    
    if (locked)
        [[NSColor redColor] set];
    else
        [[NSColor knobColor] set];
    
    [selectPath fill];
    [knobTransform invert];
    [knobTransform concat];
}

- (void) drawFocusRing
{
	[transform concat];
	
    NSBezierPath *drawPath = [NSBezierPath bezierPathWithRect: NSMakeRect(-width/2.0,-height/2.0,width,height)];
    [[NSColor keyboardFocusIndicatorColor] set];
	[drawPath setLineWidth:2];
    [drawPath stroke];
	
}

- (bool) intersectsRect: (NSRect) rect
{    
    NSBezierPath *path = [NSBezierPath bezierPathWithRect:NSMakeRect(-width/2.0,-height/2.0,width,height)];
    [path transformUsingAffineTransform: transform];
    return [path intersectsRect:rect];
}

- (int) hitTest: (NSPoint)aPoint isSelected: (bool) isSelected
{
    float length = 6+2; // +2 wegen unschrfe
    
    if ((abs(width) < 16) || (abs(height) < 16)) // kleine knobs
    {
        length = 4+2;   
    }    
    
    if (isSelected)
    {
        // lower left knob
        NSRect rect = NSMakeRect(floor(1-width/2.0-length/2.0), floor(-1-height/2.0-length/2.0),length,length);
        //+1,-1 weil wir den schatten verwenden. dieser gibt die mousposition besser an
        NSBezierPath *cornerPath = [NSBezierPath bezierPathWithRect:rect];
        [cornerPath transformUsingAffineTransform: transform];
        if ([cornerPath containsPoint:aPoint])
            return LowerLeftKnob;
        
        // middle left knob
        rect.origin.y += height/2.0;
        if (!isSquare)
        {
            cornerPath = [NSBezierPath bezierPathWithRect:rect];
            [cornerPath transformUsingAffineTransform: transform];
            if ([cornerPath containsPoint:aPoint])
                return MiddleLeftKnob;
        }
        
        // upper left knob
        rect.origin.y += height/2.0;
        cornerPath = [NSBezierPath bezierPathWithRect:rect];
        [cornerPath transformUsingAffineTransform: transform];
        if ([cornerPath containsPoint:aPoint])
            return UpperLeftKnob;
        
        // upper middel knob
        rect.origin.x += width/2.0;
        if (!isSquare)
        {
            cornerPath = [NSBezierPath bezierPathWithRect:rect];
            [cornerPath transformUsingAffineTransform: transform];
            if ([cornerPath containsPoint:aPoint])
                return UpperMiddleKnob;
        }
        
        // upper right knob
        rect.origin.x += width/2.0;
        cornerPath = [NSBezierPath bezierPathWithRect:rect];
        [cornerPath transformUsingAffineTransform: transform];
        if ([cornerPath containsPoint:aPoint])
            return UpperRightKnob;
        
        // middle right knob
        rect.origin.y -= height/2.0;
        if (!isSquare)
        {
            cornerPath = [NSBezierPath bezierPathWithRect:rect];
            [cornerPath transformUsingAffineTransform: transform];
            if ([cornerPath containsPoint:aPoint])
                return MiddleRightKnob;
        }
        
        // lower right knob
        rect.origin.y -= height/2.0;
        cornerPath = [NSBezierPath bezierPathWithRect:rect];
        [cornerPath transformUsingAffineTransform: transform];
        if ([cornerPath containsPoint:aPoint])
            return LowerRightKnob;
        
        // lower middle knob
        rect.origin.x -= width/2.0;
        if (!isSquare)
        {
            cornerPath = [NSBezierPath bezierPathWithRect:rect];
            [cornerPath transformUsingAffineTransform: transform];
            if ([cornerPath containsPoint:aPoint])
                return LowerMiddleKnob;
        }
    }
    
    // frame hit
    NSBezierPath *framePath = [NSBezierPath bezierPathWithRect:NSMakeRect(-width/2.0,-height/2.0,width,height)];
    [framePath transformUsingAffineTransform: transform];
    if ([framePath containsPoint:aPoint])
    {
        return Object;
    }
    
    return NoKnob; 
}

- (enum corner) movingKnob: (enum corner) knob fromPoint: (NSPoint) lastPoint toPoint: (NSPoint) point 
				 withEvent: (NSEvent *) theEvent
{
    if ((locked) &&	!(([undoManager isUndoing]) || ([undoManager isRedoing])))
        return knob;
    
	BOOL control = (([theEvent modifierFlags] & NSControlKeyMask) ? YES : NO);
	BOOL shift = (([theEvent modifierFlags] & NSShiftKeyMask) ? YES : NO);
	BOOL alternate = (([theEvent modifierFlags] & NSAlternateKeyMask) ? YES : NO);
	//BOOL command = (([theEvent modifierFlags] & NSCommandKeyMask) ? YES : NO);
	
    NSAffineTransform *invertTrans = [transform copy];
    [invertTrans invert];
	
	if (control) // rotieren
	{
		// prepare point
		point = [invertTrans transformPoint:point];
		lastPoint = [invertTrans transformPoint:lastPoint];	

		if (shift) // schrittweise
			[self rotateKnob: knob fromPoint: lastPoint toPoint: point stepwise: YES];
		else
			[self rotateKnob: knob fromPoint: lastPoint toPoint: point stepwise: NO];
	}
	else 
	{	
		if (isSquare)
		{
			shift = true;
			alternate = false;
		}
		NSRect newRect;
		double xOffset,yOffset;
		
		// prepare point
		point = [invertTrans transformPoint:point];
		lastPoint = [invertTrans transformPoint:lastPoint];			
		xOffset = point.x - lastPoint.x;
		yOffset = point.y - lastPoint.y;
		
		if ((shift) && (!alternate))
		{
			newRect = [self resizeWithFixedProportionsKnob: &knob xOffset: xOffset yOffset: yOffset];
		}
		else if ((alternate) && (!shift))
		{
			newRect = [self resizeAroundCenterKnob: knob xOffset: xOffset yOffset: yOffset];
		}
		else
		{
			newRect = [self resizeKnob: knob xOffset: xOffset yOffset: yOffset];
		}
		
		// knobs anpassen
		if (newRect.size.width < 0.0) 
		{
			knob = [OCEObject flipKnob:knob horizontal:YES];
			newRect.size.width = abs(newRect.size.width);
			newRect.origin.x -= newRect.size.width;
		}		
		// knobs anpassen
		if (newRect.size.height < 0.0)
		{
			knob = [OCEObject flipKnob:knob horizontal:NO];
			newRect.size.height = abs(newRect.size.height);
			newRect.origin.y -= newRect.size.height;
		}
		[self resizeTo:newRect];
	}
		
    [self reset];
    return knob;
}

- (void) moveXBy: (double) moveX moveYBy: (double) moveY;
{
	if ((locked) &&	!(([undoManager isUndoing]) || ([undoManager isRedoing])))
        return;

	xPos += moveX;
	yPos += moveY;
	[self reset];
}

- (void) startManipulationWithSelector:(SEL)aSelector andData: (id) data;
{
	offsetRect = NSZeroRect;
    saveUndoValuesArray[0] = xPos;
    saveUndoValuesArray[1] = yPos;
    saveUndoValuesArray[2] = width;
    saveUndoValuesArray[3] = height;
    saveUndoValuesArray[4] = rotation;
    [super startManipulationWithSelector: aSelector andData: data];    
}

- (void) stopManipulation
{
	if (!editingObject)
		return;
	
	offsetRect = NSZeroRect;

    bool didChange = false;
    if (saveUndoValuesArray[0] != xPos)
        didChange = true;
    
    if (saveUndoValuesArray[1] != yPos)
        didChange = true;
    
    if (saveUndoValuesArray[2] != width)
        didChange = true;
    
    if (saveUndoValuesArray[3] != height)
        didChange = true;
    
    if (saveUndoValuesArray[4] != rotation)
        didChange = true;
        
    if (didChange)
    {
        [[undoManager prepareWithInvocationTarget:self] resetWithX: saveUndoValuesArray[0] y: saveUndoValuesArray[1] 
                                                             width: saveUndoValuesArray[2] height: saveUndoValuesArray[3] 
                                                          rotation: saveUndoValuesArray[4]];
		if ([NSStringFromSelector(manipulationSelector) isEqualToString: @"moveXBy:moveYBy:"])
			[undoManager setActionName: @"Move"];
		else if ([NSStringFromSelector(manipulationSelector) isEqualToString: @"movingKnob:fromPoint:toPoint:withEvent:"])
			[undoManager setActionName: @"Resize"];

    }
	
	
	[super stopManipulation];   
}

- (void) gridObject: (int) gridSize
{
	float w = width;
	float h = height;
	
	// width
	int faktor = floor(w / gridSize);
    float offset = w - (faktor*gridSize);
    if (offset <= (gridSize/2.0))
        w = faktor*gridSize;
    else
        w = (faktor+1)*gridSize;
	
	//height
	faktor = floor(h / gridSize);
    offset = h - (faktor*gridSize);
    if (offset <= (gridSize/2.0))
        h = faktor*gridSize;
    else
        h = (faktor+1)*gridSize;
	
	if (isSquare)
	{
		if (w > h)
			[self resizeTo: NSMakeRect(xPos,yPos,w,w)];
		else
			[self resizeTo: NSMakeRect(xPos,yPos,h,h)];
	}
	else
	{
		[self resizeTo: NSMakeRect(xPos,yPos,w,h)];
	}
	[self reset];

	// grid origin
	NSPoint point = [OCEObject gridPoint: [self origin] withGridSize: gridSize];
	[self setOrigin: point];

}

- (NSPoint) pointForKnob: (enum corner) knob
{
	NSPoint point = NSZeroPoint;
	// die middleknobs sind oft == 0 deshalb mssen sie nicht extra gesetzt werden
	
	if ((knob == LowerLeftKnob) || (knob == LowerMiddleKnob) || (knob == LowerRightKnob))
		point.y = - height/2.0;
	
	if ((knob == UpperLeftKnob) || (knob == UpperMiddleKnob) || (knob == UpperRightKnob))
		point.y = height/2.0;
	
	if ((knob == UpperLeftKnob) || (knob == MiddleLeftKnob) || (knob == LowerLeftKnob))
		point.x = -width/2.0;
		
	if ((knob == UpperRightKnob) || (knob == MiddleRightKnob) || (knob == LowerRightKnob))
		point.x = width/2.0;
	
	return [transform transformPoint: point];
}

- (OCERectObject *) checkIfFocusRingObject: (NSPoint)point isStartPoint:(bool) startPoint
{
	if ([self hitTest:point isSelected: NO])
		return self;
	else
		return nil;
}

// ----------------------------------------------------------------------
// Set Methods
// ----------------------------------------------------------------------
- (void) resizeTo: (NSRect) rect
{
    double theta = rotation * M_PI / 180;
    double cosTheta = cos(theta);
    double sinTheta = sin(theta);
	
    double widthOffset = (rect.size.width-width)/2.0;
    double heightOffset = (rect.size.height-height)/2.0;
    double xOffset = rect.origin.x-xPos;
	double yOffset = rect.origin.y-yPos;
	
    height = rect.size.height;
    width = rect.size.width;
    xPos += cosTheta * xOffset + sinTheta * yOffset;
    yPos += sinTheta * xOffset - cosTheta * yOffset;
    xPos += cosTheta * widthOffset + sinTheta * heightOffset;
    yPos += sinTheta * widthOffset - cosTheta * heightOffset;
}

- (float) rotation
{
	return rotation;
}

- (void) setRotation: (float) newRotation
{
    rotation = fmod(newRotation,360);
	
	if (rotation < 0)
		rotation = 360-rotation;
	[self reset];
}

- (void) resetWithX: (double) x y: (double) y width: (double) w height: (double) h rotation: (float) r
{
    // first save old values for redo
    [[undoManager prepareWithInvocationTarget:self] resetWithX: xPos y: yPos  width: width height: height 
                                                      rotation: rotation];    
    // then set undo values and rebuild transformation
    xPos = x;
    yPos = y;
    width = w;
    height = h;

    rotation = r;
    [self reset];
}

- (bool) isSquare
{
    return isSquare;
}

- (double) width
{
    return width;
}

- (void) setWidth: (double) w
{
	if ((locked) &&	!(([undoManager isUndoing]) || ([undoManager isRedoing])))
        return;

	if (isSquare)
	{
		if ((w != width) || (w != height))
		{
			[[undoManager prepareWithInvocationTarget:self] resetWithX: xPos y: yPos  width: width height: height 
															  rotation: rotation];
			[undoManager setActionName: @"Resize"];
			[self resizeTo: NSMakeRect(xPos,yPos,w,w)];
			[self reset];
		}
	}
	else
	{
		if (w != width)
		{
			[[undoManager prepareWithInvocationTarget:self] resetWithX: xPos y: yPos  width: width height: height 
															  rotation: rotation];
			[undoManager setActionName: @"Resize"];
			[self resizeTo: NSMakeRect(xPos,yPos,w,height)];
			[self reset];
		}
	}
}

- (double) height
{
    return height;
}

- (void) setHeight: (double) h
{
    if ((locked) &&	!(([undoManager isUndoing]) || ([undoManager isRedoing])))
        return;

	double offset = h-height;
	
	if (isSquare)
	{
		if ((h != height) || (h != width))
		{			
			[[undoManager prepareWithInvocationTarget:self] resetWithX: xPos y: yPos  width: width height: height 
																	  rotation: rotation];
			[undoManager setActionName: @"Resize"];
			[self resizeTo: NSMakeRect(xPos,yPos-offset,h,h)];
			[self reset];
		}
	}
	else
	{
		if (h != height)
		{
			[[undoManager prepareWithInvocationTarget:self] resetWithX: xPos y: yPos  width: width height: height 
															  rotation: rotation];
			[undoManager setActionName: @"Resize"];
			[self resizeTo: NSMakeRect(xPos,yPos-offset,width,h)];
			[self reset];
		}
	}
}

- (NSPoint) origin
{
    return [transform transformPoint:NSMakePoint(-width/2.0,height/2.0)];   
}

- (void)  setOrigin: (NSPoint) aPoint
{
	if ((locked) &&	!(([undoManager isUndoing]) || ([undoManager isRedoing])))
		return;
	NSPoint origin = [self origin];

	if (!NSEqualPoints(aPoint, origin))
	{
		[[undoManager prepareWithInvocationTarget:self] resetWithX: xPos y: yPos  width: width height: height 
														  rotation: rotation];
		[undoManager setActionName: @"Resize"];

		xPos += aPoint.x - origin.x;
		yPos += aPoint.y - origin.y;
		[self reset];
	}
}

- (double) originX
{
    return [self origin].x;    
}

- (void) setOriginX: (double) x
{
	[self setOrigin: NSMakePoint(x,[self origin].y)];
}

- (double) originY
{
    return [self origin].y;    
}

- (void) setOriginY: (double) y
{
	[self setOrigin: NSMakePoint([self origin].x,y)];
}

- (NSPoint) center
{
	return NSMakePoint(NSMidX(bounds),NSMidY(bounds)); 
}

- (NSBezierPath *) path
{
	NSBezierPath *framePath = [NSBezierPath bezierPathWithRect:NSMakeRect(-width/2.0,-height/2.0,width,height)];
    [framePath transformUsingAffineTransform: transform];
	return framePath;
}

// ----------------------------------------------------------------------
// manipulate
// ----------------------------------------------------------------------
- (void) rotateKnob: (enum corner)knob fromPoint: (NSPoint)lastPoint toPoint: (NSPoint)point stepwise: (bool)flag
{
	float newRotation = [OCEMath calculateDegreesVector1: (NSPoint) {1,0} vector2: point];
	
	NSPoint knobPoint;
	if (knob == LowerLeftKnob)
	{
		knobPoint.x = -width/2.0;
		knobPoint.y = -height/2.0;
	}
	else if (knob == UpperLeftKnob)
	{
		knobPoint.x = -width/2.0;
		knobPoint.y = height/2.0;
	}
	else if (knob == UpperRightKnob)
	{
		knobPoint.x = width/2.0;
		knobPoint.y = height/2.0;
	}
	else if (knob == LowerRightKnob) 
	{
		knobPoint.x = width/2.0;
		knobPoint.y = -height/2.0;
	}
	else if (knob == MiddleLeftKnob)
	{
		knobPoint.x = -width/2.0;
		knobPoint.y = 0;
	}
	else if (knob == UpperMiddleKnob) 
	{
		knobPoint.x = 0;
		knobPoint.y = height/2.0;
	}        
	else if (knob == MiddleRightKnob)
	{
		knobPoint.x = width/2.0;
		knobPoint.y = 0;
	}
	else if (knob == LowerMiddleKnob) 
	{
		knobPoint.x = 0;
		knobPoint.y = -height/2.0;
	}
	
	point = [transform transformPoint:point];
	float knobRotation = [OCEMath calculateDegreesVector1: (NSPoint) {1,0} vector2: knobPoint];
	newRotation -= knobRotation;
	newRotation += rotation;
	
	if (flag)
	{
		int times = newRotation / 15;
		float remainder = (int)newRotation % 15;
		if (remainder <= 7.5)
			newRotation = times*15;
		else 
			newRotation = (times+1)*15;
	}
	
	[self setRotation: newRotation];
}

- (NSRect) resizeKnob: (enum corner)knob xOffset: (float)xOffset yOffset: (float)yOffset
{
	NSRect newRect = NSMakeRect(xPos+offsetRect.origin.x,yPos+offsetRect.origin.y,
								width+offsetRect.size.width,height+offsetRect.size.height);
	offsetRect = NSZeroRect;
	
	if ((knob == UpperLeftKnob) || (knob == MiddleLeftKnob) || (knob == LowerLeftKnob)) 
	{
		// Adjust left edge
		newRect.size.width -= xOffset;
		newRect.origin.x += xOffset;
	} 
	else if ((knob == UpperRightKnob) || (knob == MiddleRightKnob) || (knob == LowerRightKnob)) 
	{
		// Adjust right edge
		newRect.size.width += xOffset;
	}
	if ((knob == UpperLeftKnob) || (knob == UpperMiddleKnob) || (knob == UpperRightKnob)) 
	{
		// Adjust top edge
		newRect.size.height += yOffset;
		newRect.origin.y -= yOffset;
	} 
	else if ((knob == LowerLeftKnob) || (knob == LowerMiddleKnob) || (knob == LowerRightKnob)) 
	{
		// Adjust bottom edge
		newRect.size.height -= yOffset;
	}
	
	return newRect;
}

- (NSRect) resizeAroundCenterKnob: (enum corner)knob xOffset: (float)xOffset yOffset: (float)yOffset
{
	NSRect newRect = NSMakeRect(xPos+offsetRect.origin.x,yPos+offsetRect.origin.y,
								width+offsetRect.size.width,height+offsetRect.size.height);
	offsetRect = NSZeroRect;
	
	if ((knob == UpperLeftKnob) || (knob == MiddleLeftKnob) || (knob == LowerLeftKnob))
	{
		newRect.origin.x += xOffset;
		newRect.size.width -= 2*xOffset;
	}
	if ((knob == UpperRightKnob) || (knob == MiddleRightKnob) || (knob == LowerRightKnob)) 
	{
		newRect.origin.x -= xOffset;
		newRect.size.width += 2*xOffset;
	}
	
	if ((knob == UpperLeftKnob) || (knob == UpperMiddleKnob) || (knob == UpperRightKnob))
	{
		newRect.origin.y -= yOffset;
		newRect.size.height += 2*yOffset;
	}
	if ((knob == LowerLeftKnob) || (knob == LowerMiddleKnob) || (knob == LowerRightKnob)) 
	{
		newRect.origin.y += yOffset;
		newRect.size.height -= 2*yOffset;
	}
	
	return newRect;
}

- (NSRect) resizeWithFixedProportionsKnob: (enum corner *)knob xOffset: (float)xOffset yOffset: (float)yOffset
{
	// resize the rect, will also set offsetRect to Zero
	NSRect newRect = [self resizeKnob: *knob xOffset:xOffset yOffset:yOffset];
	double absWidth = abs(newRect.size.width);
	double absHeight = abs(newRect.size.height);
	double proportionHeight = absHeight / proportion;
	
	if ((*knob == MiddleRightKnob) || (*knob == MiddleLeftKnob))
	{	// resize at the left/right side
		if (newRect.size.width < 0.0) 
		{ // adapt knob
			*knob = [OCEObject flipKnob:*knob horizontal:YES];
			newRect.size.width = abs(newRect.size.width);
			newRect.origin.x -= newRect.size.width;
		}
		double oldHeight = newRect.size.height;
		newRect.size.height = newRect.size.width * proportion;
		newRect.origin.y -= newRect.size.height - oldHeight; // damit die hhe nach oben wchst und nicht nach unten
	}
	else if ((*knob == UpperMiddleKnob) || (*knob == LowerMiddleKnob))
	{ // upper/lower side
		if (newRect.size.height < 0.0)
		{ // adapt knob
			*knob = [OCEObject flipKnob:*knob horizontal:NO];
			newRect.size.height = abs(newRect.size.height);
			newRect.origin.y -= newRect.size.height;
		}
		newRect.size.width = newRect.size.height / proportion;
	}
	else if (proportionHeight > absWidth)
	{ // height anpassen
		double proportionWidth = absWidth * proportion;
		if ((newRect.size.height < 0.0) && (newRect.size.width < 0.0))
		{
			// hier height anpassen. width wird ausserhalb diese methode angepasst
			offsetRect.size.height = absHeight-proportionWidth;
			newRect.origin.y -= absHeight;
			*knob = [OCEObject flipKnob: *knob horizontal:NO];
		}
		else
		{
			offsetRect.size.height = newRect.size.height-proportionWidth;
			if ((*knob == UpperLeftKnob) || (*knob == UpperRightKnob)) 
			{
				newRect.origin.y += newRect.size.height-proportionWidth;
				offsetRect.origin.y = -offsetRect.size.height;
			}
		}
		newRect.size.height = proportionWidth;
	}
	else if (absWidth > proportionHeight)
	{
		if ((newRect.size.height < 0.0) && (newRect.size.width < 0.0))
		{
			// hier height anpassen. width wird ausserhalb diese methode angepasst
			offsetRect.size.width = absWidth-proportionHeight;
			newRect.origin.x -= absWidth;
			*knob = [OCEObject flipKnob: *knob horizontal:YES];
		}
		else
		{
			offsetRect.size.width = newRect.size.width-proportionHeight;
			if ((*knob == UpperLeftKnob) || (*knob == LowerLeftKnob)) 
			{
				newRect.origin.x += newRect.size.width-proportionHeight;
				offsetRect.origin.x = -offsetRect.size.width;
			}
		}
		newRect.size.width = proportionHeight;
	}
	
	return newRect;
}

- (void) reset
{
	// because we call this method everytime we change the rect we can do some checking here
	// check if width or height == 0
	//width = ((width == 0) ? 1 : width);
	//height = ((height == 0) ? 1 : height);

    [transform release];
	transform = [[NSAffineTransform transform] retain];
    [transform rotateByDegrees: rotation];

    NSAffineTransform *translateTransform = [NSAffineTransform transform];
    [translateTransform translateXBy: xPos yBy: yPos];
    [transform appendTransform:translateTransform];
    
    // set bounds
    // -3-1=-4 damit auch die knobs korrekt gezeichnet werden
    NSBezierPath *framePath = [NSBezierPath bezierPathWithRect:NSMakeRect(-width/2.0-4,-height/2.0-4,width+8,height+8)];
    [framePath transformUsingAffineTransform: transform];
    bounds = [framePath bounds];
	
	// associatedPaths
	NSEnumerator *enumerator = [associatedPaths objectEnumerator];
	OCEPathObject *object;
	
	while (object = [enumerator nextObject]) 
	{
		[object adaptPathToRects];
	}
}

- (void) addedToCanvas: (OCEDocument *) document
{
	[super addedToCanvas: document];
	if (rectID == 0)
		[self setRectID:[document newRectObjectID]]; // get new id
}

- (void) removedFromCanvas: (OCEDocument *) document
{
	[super removedFromCanvas: document];
	[self unregisterAssociatedPaths];
}

// ----------------------------------------------------------------------
// NSCopying Protocol
// ----------------------------------------------------------------------
- (id)copyWithZone:(NSZone *)zone
{
    OCERectObject *copy = [super copyWithZone: zone];
    
    copy->xPos = xPos+10;
    copy->yPos = yPos+10;
    copy->width = width;
    copy->height = height;
    copy->rotation = rotation;
    copy->isSquare = isSquare;
    copy->offsetRect = NSZeroRect;
    [copy reset];
    
    return copy;
}

// ----------------------------------------------------------------------
// save/load stuff
// ----------------------------------------------------------------------
- (NSMutableArray *)dataRepresentation
{
	NSMutableArray *array = [[super dataRepresentation] retain];
	[array addObject: [NSNumber numberWithFloat:xPos]];
	[array addObject: [NSNumber numberWithFloat:yPos]];
	[array addObject: [NSNumber numberWithFloat:width]];
	[array addObject: [NSNumber numberWithFloat:height]];
	[array addObject:[NSNumber numberWithFloat: rotation]];
	[array addObject:[NSNumber numberWithInt: rectID]];
		
    return [array autorelease];
}

- (id)initWithContentsOfArray:(NSArray *) array forDocument: (OCEDocument *) newDocument
{
	xPos = [[array objectAtIndex:2] floatValue];
	yPos = [[array objectAtIndex:3] floatValue];
	width = [[array objectAtIndex:4] floatValue];
	height = [[array objectAtIndex:5] floatValue];
	rotation = [[array objectAtIndex:6] floatValue];
	rectID = [[array objectAtIndex:7] intValue];

	[self reset];

    return [super initWithContentsOfArray: array forDocument: newDocument];
}


// ----------------------------------------------------------------------
// associated paths
// ----------------------------------------------------------------------
- (void) addAssociatedPath: (OCEPathObject *) path // doesn't register the trail
{
	[associatedPaths addObject:path];
}

- (void) removeAssociatedPath: (OCEPathObject *) path  // doesn't unregister the trail
{
	// es kann ein path 2x registriert sein (Start,EndPoint). diese methode lscht den pfad nur 1x
	int index = [associatedPaths indexOfObjectIdenticalTo:path];
	if (index != NSNotFound)
		[associatedPaths removeObjectAtIndex:index];
}

- (void) unregisterAssociatedPaths
{
	int i = [associatedPaths count]-1;
	while (i > -1)
	{
		OCEPathObject *object = [associatedPaths objectAtIndex:i];
		if (((signed)[associatedPaths indexOfObjectIdenticalTo:object]) != i)
			i--; // start und end Point zeigen auf dieses Object
		
		[object unregisterAssociatedRect:self];
		i--;
	}
}

- (NSArray *) outgoingAssociatedPaths
{
	NSMutableArray *array = [[NSMutableArray arrayWithCapacity:0] retain];
	
	int i = 0;
	int c = [associatedPaths count];
	
	while (i < c)
	{
		OCEPathObject *object = [associatedPaths objectAtIndex:i];
		
		if (([object associatedStartRect] == self) && ([array indexOfObjectIdenticalTo: object] == NSNotFound))
			[array addObject:object];
		
		i++;
	}
	
	return [array autorelease];
}

// ----------------------------------------------------------------------
// ID Methoden
// ----------------------------------------------------------------------
- (int) rectID
{
    return rectID;
}

- (void) setRectID: (int) newRectID
{
    rectID = newRectID;
}

- (OCERectObject *) rectObjectFromID: (int) anID;
{
	if (anID == rectID)
		return self;
	else
		return nil;
}

// ----------------------------------------------------------------------
// OTHER
// ----------------------------------------------------------------------
- (bool) isActiveComponent;
{
	return NO;	
}

@end
